APP_PATH = "../tofuref/main.py"


def test_welcome(snap_compare):
    assert snap_compare(APP_PATH, terminal_size=(200, 60))


def test_welcome_fullscreen(snap_compare):
    assert snap_compare(APP_PATH)


def test_toggle_fullscreen(snap_compare):
    assert snap_compare(APP_PATH, terminal_size=(200, 30), press="f")


def test_content(snap_compare):
    assert snap_compare(APP_PATH, press=["c", "pagedown"])


def test_search_github(snap_compare):
    assert snap_compare(APP_PATH, press=["s", "g", "i", "t", "h", "u", "b"])


def test_open_github(snap_compare):
    assert snap_compare(
        APP_PATH, press=["s", "g", "i", "t", "h", "u", "b", "enter", "enter"]
    )


def test_open_github_membership(snap_compare):
    assert snap_compare(
        APP_PATH,
        press=[
            "s",
            "g",
            "i",
            "t",
            "h",
            "u",
            "b",
            "enter",
            "enter",
            "s",
            "m",
            "e",
            "m",
            "b",
            "e",
            "r",
            "enter",
            "enter",
        ],
    )


def test_back_to_providers(snap_compare):
    assert snap_compare(APP_PATH, press=["enter", "p"])


def test_provider_overview(snap_compare):
    assert snap_compare(
        APP_PATH, press=["s", "g", "i", "t", "h", "u", "b", "enter", "enter", "c"]
    )


def test_version_picker(snap_compare):
    assert snap_compare(
        APP_PATH, press=["s", "g", "i", "t", "h", "u", "b", "enter", "enter", "v"]
    )


def test_version_picker_submit(snap_compare):
    assert snap_compare(
        APP_PATH,
        press=[
            "s",
            "g",
            "i",
            "t",
            "h",
            "u",
            "b",
            "enter",
            "enter",
            "v",
            "down",
            "enter",
        ],
    )


def test_use(snap_compare):
    assert snap_compare(
        APP_PATH, press=["s", "g", "i", "t", "h", "u", "b", "enter", "enter", "u"]
    )



from textual.binding import Binding
from textual.widgets import (
    Input,
    OptionList,
    RichLog,
    MarkdownViewer,
)

from tofuref import __version__


class CustomRichLog(RichLog):
    def __init__(self, **kwargs):
        super().__init__(
            id="log", markup=True, wrap=True, classes="bordered hidden", **kwargs
        )
        self.border_title = "Log"
        self.border_subtitle = f"tofuref v{__version__}"
        self.display = False


class WelcomeMarkdownViewer(MarkdownViewer):
    ALLOW_MAXIMIZE = True

    BINDINGS = [
        Binding("up", "up", "Scroll Up", show=False),
        Binding("down", "down", "Scroll Down", show=False),
        Binding("pageup", "page_up", "Page Up", show=False),
        Binding("pagedown", "page_down", "Page Down", show=False),
        Binding("home", "scroll_home", "Top", show=False),
        Binding("end", "scroll_end", "Bottom", show=False),
    ]

    def __init__(self, content=None, **kwargs):
        welcome_content = f"""
# Welcome to tofuref {__version__}!

Changelog: https://github.com/djetelina/tofuref/blob/main/CHANGELOG.md

## Controls
Navigate with arrows/page up/page down/home/end or your mouse.

| keybindings | action |
|------|--------|
| `tab` | focus next window |
| `shift+tab` | focus previous window |
| `enter` | choose selected or finish search |
| `q`, `ctrl+q` | **quit** tofuref |
| `s`, `/` | **search** in the context of providers and resources |
| `v` | change active provider **version** |
| `p` | focus **providers** window |
| `u`, `y` | copy ready-to-**use** provider version constraint |
| `r` | focus **resources** window |
| `c` | focus **content** window |
| `f` | toggle **fullscreen** mode |
| `l` | display **log** window |

---

# Get in touch
* GitHub: https://github.com/djetelina/tofuref"""

        super().__init__(
            content if content is not None else welcome_content,
            classes="content",
            show_table_of_contents=False,
            id="content",
            **kwargs,
        )

    def action_up(self) -> None:
        self.document.scroll_up()

    def action_down(self) -> None:
        self.document.scroll_down()

    def action_page_down(self) -> None:
        self.document.action_page_down()

    def action_page_up(self) -> None:
        self.document.action_page_up()

    def action_scroll_home(self) -> None:
        self.document.scroll_home()

    def action_scroll_end(self) -> None:
        self.document.scroll_end()

    # Without this, the Markdown viewer would try to open a file on a disk, while the Markdown itself will open a browser link (desired)
    async def go(self, location):
        return None


class SearchInput(Input):
    def __init__(self, **kwargs):
        super().__init__(
            placeholder="Search...", id="search", classes="bordered", **kwargs
        )
        self.border_title = "Search"


class ProvidersOptionList(OptionList):
    def __init__(self, **kwargs):
        super().__init__(
            name="Providers",
            id="nav-provider",
            classes="nav-selector bordered",
            **kwargs,
        )
        self.border_title = "Providers"


class ResourcesOptionList(OptionList):
    def __init__(self, **kwargs):
        super().__init__(
            name="Resources",
            id="nav-resources",
            classes="nav-selector bordered",
            **kwargs,
        )
        self.border_title = "Resources"



import re
from typing import Tuple, Union, Dict, Any, Optional
import logging

import httpx
from async_lru import alru_cache
from yaml import safe_load
from yaml.scanner import ScannerError

from tofuref import __version__

LOGGER = logging.getLogger(__name__)


def header_markdown_split(contents: str) -> Tuple[dict, str]:
    """
    Most of the documentation files from the registry have a YAML "header"
    that we mostly (at the moment) don't care about. Either way we
    check for the header and if it's there, we split it.
    """
    header = {}
    if re.match(r"^---$", contents, re.MULTILINE):
        split_contents = re.split(r"^---$", contents, 3, re.MULTILINE)
        try:
            header = safe_load(split_contents[1])
        except ScannerError as _:
            header = {}
        markdown_content = split_contents[2]
    else:
        markdown_content = contents
    return header, markdown_content


@alru_cache(maxsize=64)
async def get_registry_api(
    endpoint: str, json: bool = True, log_widget: Optional[Any] = None
) -> Union[Dict[str, dict], str]:
    """
    Sends GET request to opentofu providers registry to a given endpoint
    and returns the response either as a JSON or as a string. It also "logs" the request.

    The function is cached because of tests. In practice, the cache will never get used.
    """
    uri = f"https://api.opentofu.org/registry/docs/providers/{endpoint}"
    LOGGER.debug("Starting async client")
    async with httpx.AsyncClient(
        headers={"User-Agent": f"tofuref v{__version__}"}
    ) as client:
        LOGGER.debug("Client started, sending request")
        try:
            r = await client.get(uri)
            LOGGER.debug("Request sent, response received")
        except Exception as e:
            LOGGER.error("Something went wrong", exc_info=e)
            if log_widget is not None:
                log_widget.write(f"Something went wrong: {e}")
            return ""

    if log_widget is not None:
        log_widget.write(f"GET [cyan]{endpoint}[/] [bold]{r.status_code}[/]")

    if json:
        return r.json()
    else:
        return r.text



from typing import Optional, List, cast, Collection

from tofuref.data.resources import Resource
from tofuref.data.providers import Provider
from tofuref.data.registry import registry
from tofuref.widgets import (
    ResourcesOptionList,
    ProvidersOptionList,
    WelcomeMarkdownViewer,
)


async def load_provider_resources(
    provider: Provider,
    navigation_resources: ResourcesOptionList = None,
    content_markdown: WelcomeMarkdownViewer = None,
):
    if navigation_resources is None or content_markdown is None:
        return

    navigation_resources.loading = True
    content_markdown.loading = True
    await provider.load_resources()
    await content_markdown.document.update(await provider.overview())
    content_markdown.document.border_subtitle = (
        f"{provider.display_name} {provider.active_version} Overview"
    )
    populate_resources(provider, navigation_resources=navigation_resources)
    navigation_resources.focus()
    navigation_resources.highlighted = 0
    content_markdown.loading = False
    navigation_resources.loading = False


def populate_providers(
    providers: Optional[Collection[str]] = None,
    navigation_providers: ProvidersOptionList = None,
) -> None:
    if navigation_providers is None:
        return

    if providers is None:
        providers = registry.providers.keys()
    providers = cast(Collection[str], providers)
    navigation_providers.clear_options()
    navigation_providers.border_subtitle = f"{len(providers)}/{len(registry.providers)}"
    for name in providers:
        navigation_providers.add_option(name)


def populate_resources(
    provider: Optional[Provider] = None,
    resources: Optional[List[Resource]] = None,
    navigation_resources: ResourcesOptionList = None,
) -> None:
    if navigation_resources is None:
        return

    navigation_resources.clear_options()
    if provider is None:
        return
    navigation_resources.border_subtitle = (
        f"{provider.organization}/{provider.name} {provider.active_version}"
    )

    if resources is None:
        for resource in provider.resources:
            navigation_resources.add_option(resource)
    else:
        navigation_resources.add_options(resources)



import asyncio
import logging

from textual import on
from textual.app import App, ComposeResult
from textual.containers import Container
from textual.widgets import (
    Footer,
    Input,
    OptionList,
    Select,
)
from rich.markdown import Markdown

from tofuref.data.providers import populate_providers
from tofuref.data.registry import registry
from tofuref.ui_logic import (
    load_provider_resources,
    populate_providers as ui_populate_providers,
    populate_resources,
)
from tofuref.widgets import (
    CustomRichLog,
    WelcomeMarkdownViewer,
    ProvidersOptionList,
    ResourcesOptionList,
    SearchInput,
)

LOGGER = logging.getLogger(__name__)


class TofuRefApp(App):
    CSS_PATH = "tofuref.tcss"
    TITLE = "TofuRef - OpenTofu Provider Reference"
    BINDINGS = [
        ("q", "quit", "Quit"),
        ("s", "search", "Search"),
        ("/", "search", "Search"),
        ("v", "version", "Provider Version"),
        ("p", "providers", "Providers"),
        ("u", "use", "Use provider"),
        ("y", "use", "Use provider"),
        ("r", "resources", "Resources"),
        ("c", "content", "Content"),
        ("f", "fullscreen", "Fullscreen Mode"),
        ("l", "log", "Show Log"),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log_widget = CustomRichLog()
        self.content_markdown = WelcomeMarkdownViewer()
        self.navigation_providers = ProvidersOptionList()
        self.navigation_resources = ResourcesOptionList()
        self.search = SearchInput()

    def compose(self) -> ComposeResult:
        # Navigation
        with Container(id="sidebar"):
            with Container(id="navigation"):
                yield self.navigation_providers
                yield self.navigation_resources

        # Main content area
        with Container(id="content"):
            yield self.content_markdown

        yield self.log_widget

        yield Footer()

    async def on_ready(self) -> None:
        LOGGER.debug("Starting on ready")
        self.log_widget.write("Populating providers from the registry API")
        self.content_markdown.document.classes = "bordered content"
        self.content_markdown.document.border_title = "Content"
        self.content_markdown.document.border_subtitle = "Welcome"
        if self.size.width < 125:
            registry.fullscreen_mode = True
        if registry.fullscreen_mode:
            self.navigation_providers.styles.column_span = 2
            self.navigation_resources.styles.column_span = 2
            self.content_markdown.styles.column_span = 2
            self.screen.maximize(self.navigation_providers)
        self.navigation_providers.loading = True
        self.screen.refresh()
        await asyncio.sleep(0.001)
        LOGGER.debug("Starting on ready done, running preload worker")
        self.app.run_worker(self._preload, name="preload")

    async def _preload(self) -> None:
        LOGGER.debug("preload start")
        registry.providers = await populate_providers(log_widget=self.log_widget)
        self.log_widget.write(
            f"Providers loaded ([cyan bold]{len(registry.providers)}[/])"
        )
        ui_populate_providers(navigation_providers=self.navigation_providers)
        self.navigation_providers.loading = False
        self.navigation_providers.highlighted = 0
        self.log_widget.write(Markdown("---"))
        LOGGER.info("Initial load complete")

    def action_search(self) -> None:
        """Focus the search input."""
        if self.search.has_parent:
            self.search.parent.remove_children([self.search])
        for searchable in [self.navigation_providers, self.navigation_resources]:
            if searchable.has_focus:
                self.search.value = ""
                searchable.mount(self.search)
                self.search.focus()
                self.search.offset = searchable.offset + (
                    0,
                    searchable.size.height - 3,
                )

    def action_use(self) -> None:
        if registry.active_provider:
            self.copy_to_clipboard(registry.active_provider.use_configuration)
            self.notify(
                registry.active_provider.use_configuration, title="Copied", timeout=10
            )

    def action_log(self) -> None:
        self.log_widget.display = not self.log_widget.display

    def action_providers(self) -> None:
        if registry.fullscreen_mode:
            self.screen.maximize(self.navigation_providers)
        self.navigation_providers.focus()

    def action_resources(self) -> None:
        if registry.fullscreen_mode:
            self.screen.maximize(self.navigation_resources)
        self.navigation_resources.focus()

    def action_content(self) -> None:
        if registry.fullscreen_mode:
            self.screen.maximize(self.content_markdown)
        self.content_markdown.document.focus()

    def action_fullscreen(self) -> None:
        if registry.fullscreen_mode:
            registry.fullscreen_mode = False
            self.navigation_providers.styles.column_span = 1
            self.navigation_resources.styles.column_span = 1
            self.content_markdown.styles.column_span = 1
            self.screen.minimize()
        else:
            registry.fullscreen_mode = True
            self.navigation_providers.styles.column_span = 2
            self.navigation_resources.styles.column_span = 2
            self.content_markdown.styles.column_span = 2
            self.screen.maximize(self.screen.focused)

    async def action_version(self) -> None:
        if registry.active_provider is None:
            self.notify(
                "Provider Version can only be changed after one is selected.",
                title="No provider selected",
                severity="warning",
            )
            return
        if self.navigation_resources.children:
            await self.navigation_resources.remove_children("#version-select")
        else:
            version_select = Select.from_values(
                (v["id"] for v in registry.active_provider.versions),
                prompt="Select Provider Version",
                allow_blank=False,
                value=registry.active_provider.active_version,
                id="version-select",
            )
            await self.navigation_resources.mount(version_select)
            await asyncio.sleep(0.1)
            version_select.action_show_overlay()

    @on(Select.Changed, "#version-select")
    async def change_provider_version(self, event: Select.Changed) -> None:
        if event.value != registry.active_provider.active_version:
            registry.active_provider.active_version = event.value
            await load_provider_resources(
                registry.active_provider,
                navigation_resources=self.navigation_resources,
                content_markdown=self.content_markdown,
            )
            await self.navigation_resources.remove_children("#version-select")

    def on_input_changed(self, event: Input.Changed) -> None:
        if event.input.id != "search":
            return

        query = event.value.strip()
        if self.search.parent == self.navigation_providers:
            if not query:
                ui_populate_providers(navigation_providers=self.navigation_providers)
            else:
                ui_populate_providers(
                    [p for p in registry.providers.keys() if query in p],
                    navigation_providers=self.navigation_providers,
                )
        elif self.search.parent == self.navigation_resources:
            if not query:
                populate_resources(
                    registry.active_provider,
                    navigation_resources=self.navigation_resources,
                )
            else:
                populate_resources(
                    registry.active_provider,
                    [r for r in registry.active_provider.resources if query in r.name],
                    navigation_resources=self.navigation_resources,
                )

    def on_input_submitted(self, event: Input.Submitted) -> None:
        self.search.parent.focus()
        self.search.parent.highlighted = 0
        self.search.parent.remove_children([self.search])

    async def on_option_list_option_selected(
        self, event: OptionList.OptionSelected
    ) -> None:
        if event.control == self.navigation_providers:
            provider_selected = registry.providers[event.option.prompt]
            registry.active_provider = provider_selected
            if registry.fullscreen_mode:
                self.screen.maximize(self.navigation_resources)
            await load_provider_resources(
                provider_selected,
                navigation_resources=self.navigation_resources,
                content_markdown=self.content_markdown,
            )
        elif event.control == self.navigation_resources:
            resource_selected = event.option.prompt
            if registry.fullscreen_mode:
                self.screen.maximize(self.content_markdown)
            self.content_markdown.loading = True
            await self.content_markdown.document.update(
                await resource_selected.content()
            )
            self.content_markdown.document.border_subtitle = f"{resource_selected.type.value} - {resource_selected.provider.name}_{resource_selected.name}"
            self.content_markdown.document.focus()
            self.content_markdown.loading = False


def main() -> None:
    LOGGER.debug("Starting tofuref")
    TofuRefApp().run()


if __name__ == "__main__":
    main()
